use std::collections::HashMap;
use crate::parse::Parser;
use crate::parse::TokenType::*;

#[derive(Debug)]
pub struct Vm {
    stack:      Vec<u64>,
    dictionary: Dictionary
}

#[derive(Debug, Default)]
struct Dictionary {
    dp:    u64,
    data:  HashMap<u64, u64>,
    words: HashMap<String, String>
}


impl<'a> Vm {
    pub fn new() -> Self {
	Self {
	    stack: Vec::new(),
	    dictionary: Default::default()
	}
    }
    
    pub fn vmrun(&mut self, parser: &mut Parser) {
	macro_rules! unwrap {
	    () => {
		match self.stack.pop() {
		    Some(v) => {
			v
		    },
		    None => {
			println!("ERROR: stack empty");
			std::process::exit(1);
		    }
		}
	    }
	}
	macro_rules! paren_matching {
	    ($start:expr, $end:expr) => {{
		let start = $start;
		let end   = $end;
		let mut depth = 1;
		let mut content = String::new();
		
		let item = {
		    loop {
			if let Some(item) = parser.inner_next() {
			    content.push_str(item);
			    content.push_str(" ");
			    if start(item) {
				depth += 1;
			    }
			    if end(item) {
				depth -= 1;
			    }
			    if depth == 0 {
				break Some(item);
			    }
			} else {
			    break None;
			}
		    }
		};
	    
		(item, content)
	    }}
	}

	while let Some(item) = parser.next() {
	    match item {
		Number(n) => self.stack.push(n),
		Word(w) => {
		    match w {
			"." => {
			    println!("{:?} ok", unwrap!());
			},
			"+" => {
			    let sum = unwrap!().wrapping_add(unwrap!());
			    self.stack.push(sum);
			},
			"*" => {
			    let prod = unwrap!().wrapping_mul(unwrap!());
			    self.stack.push(prod);
			},
			"/" => {
			    let stack_top = unwrap!();
			    let stack_second = unwrap!();
			    let quo = stack_second.wrapping_div(stack_top);
			    self.stack.push(quo);
			},
			"=" => {
			    if unwrap!() == unwrap!() {
				self.stack.push(1);
			    } else {
				self.stack.push(0);
			    }
			},
			"<>" => {
			    if unwrap!() != unwrap!() {
				self.stack.push(1);
			    } else {
				self.stack.push(0)
			    }
			},
			"<" => {
			    if unwrap!() > unwrap!() {
				self.stack.push(1);
			    } else {
				self.stack.push(0);
			    }
			},
			">" => {
			    if unwrap!() < unwrap!() {
				self.stack.push(1);
			    } else {
				self.stack.push(0);
			    }
			}
			"MOD"|"mod" => {
			    let stack_top = unwrap!();
			    let stack_second = unwrap!();
			    let rem = stack_second.wrapping_rem(stack_top);
			    self.stack.push(rem);
			},
			"*/" => {
			    let n3 = unwrap!();
			    let n2 = unwrap!();
			    let n1 = unwrap!();

			    let d = n1 * n2;
			    let n4 = d/n3;

			    self.stack.push(n4);
			},
			"TRACE"|"trace" => {
			    match parser.inner_next() {
				Some("stack") => {
				    for item in &self.stack {
					print!("{:?} ", item);
				    }
				    println!("<- TOP");
				},
				Some("data") => {
				    println!("{:?}", self.dictionary.data);
				},
				Some("words") => {
				    println!("{:?}", self.dictionary.words);
				},

				subcommand => {
				    println!("Subcommand {:?} is not supported. Supported subcommands are stack or dataspace.", subcommand); 
				}
			    }
			},
			"/MOD"|"/mod" => {
			    let stack_top = unwrap!();
			    let stack_second = unwrap!();
			    let rem = stack_second.wrapping_rem(stack_top);
			    let quo = stack_second.wrapping_div(stack_top);
			    self.stack.push(rem);
			    self.stack.push(quo);
			},
			"SWAP"|"swap" => {
			    let stack_top = unwrap!();
			    let stack_second = unwrap!();
			    self.stack.push(stack_top);
			    self.stack.push(stack_second);
			},
			"DROP"|"drop" => {
			    let _ = self.stack.pop();
			},
			"HERE"|"here" => {
			    self.stack.push(self.dictionary.dp);
			},
			"@" => {
			    let addr = unwrap!();
			    self.stack.push(*self.dictionary.data.get(&addr).unwrap_or(&0));
			},
			"!" => {
			    let addr = unwrap!();
			    let content = unwrap!();
			    self.dictionary.data.insert(addr, content);
			},
			"," => {
			    let content = unwrap!();
			    self.dictionary.dp += 1;
			    self.dictionary.data.insert(self.dictionary.dp, content);
			},
			"ALLOT"|"allot" => {
			    self.dictionary.dp += unwrap!();
			},
			"CREATE"|"create" => {
			    self.dictionary.dp += 1;
			    let Some(Word(word_name)) = parser.next() else { println!("ERROR: Word name not given."); std::process::exit(1) };
			    self.dictionary.words.insert(word_name.to_string(), self.dictionary.dp.to_string());
			},
			"VARIABLE"|"variable" => {
			    self.dictionary.dp += 1;
			    let Some(Word(word_name)) = parser.next() else { println!("ERROR: Word name not given."); std::process::exit(1) };
			    self.dictionary.words.insert(word_name.to_string(), self.dictionary.dp.to_string());
			    
			},
			"DUP"|"dup" => {
			    let stack_top = unwrap!();
			    self.stack.push(stack_top);
			    self.stack.push(stack_top);
			},
			"IF"|"if" => {
			    let expr = unwrap!();
			    let (_, content) = paren_matching!(|x| matches!(x, "IF"|"if"), |x| matches!(x, "THEN"|"then"));

			    if expr == 0 {
				// Do nothing.
			    } else {
				self.vmrun(&mut Parser::new(content.as_str()));
			    }
			},

			"THEN"|"then" => {}
			"(" => {
			    let mut next = parser.inner_next();
			    while next != Some(")") && next != None {
				next = parser.inner_next_back();
			    }
			},
			"stacklen" => {
			    self.stack.push(self.stack.len() as u64);
			}
			"DO"|"do" => {
			    let start = unwrap!();
			    let limit = unwrap!();
			    let (_, action) = paren_matching!(|x| matches!(x, "DO"|"do"), |x| matches!(x, "LOOP"|"loop"));

			    for i in start..limit {
				self.dictionary.words.insert("I".to_string(), i.to_string());
				self.dictionary.words.insert("i".to_string(), i.to_string());
				self.vmrun(&mut Parser::new(action.as_str()));
			    }
			    self.dictionary.words.remove("I");
			    self.dictionary.words.remove("i");
			},
			"+!" => {
			    let addr = unwrap!();
			    let amnt = unwrap!();
			    self.dictionary.data.insert(addr, self.dictionary.data.get(&addr).unwrap_or(&0) + amnt);
			},
			"LOOP"|"loop" => {},
			"BEGIN"|"begin" => {
			    let (next, action) = paren_matching!(|x| matches!(x, "BEGIN"|"begin"), |x| matches!(x, "UNTIL"|"until"|"WHILE"|"while"));
			    
			    match next {
				Some("UNTIL")|Some("until") => {
				    self.vmrun(&mut Parser::new(action.as_str()));
				    
				    while unwrap!() == 0 {
					self.vmrun(&mut Parser::new(action.as_str()));
				    }
				},
				Some("WHILE")|Some("while") => {
				    let mut condition = String::new();
				    let mut next      = parser.inner_next();
				    while !matches!(next, Some("REPEAT")|Some("repeat")|None) {
					condition.push_str(next.unwrap());
					condition.push_str(" ");
					next = parser.inner_next_back();
				    }
				    self.vmrun(&mut Parser::new(condition.as_str()));
				    let mut cond_bool = unwrap!();
				    while cond_bool != 0 {
					self.vmrun(&mut Parser::new(action.as_str()));
					self.vmrun(&mut Parser::new(condition.as_str()));
					cond_bool = unwrap!();
				    }
				},
				_ => {
				    println!("ERROR: BEGIN without UNTIL or WHILE");
				    std::process::exit(1);
				}
			    }
			},
			"UNTIL"|"until"   => {},
			"WHILE"|"while"   => {},
			"REPEAT"|"repeat" => {},
			":" => {
			    let Some(Word(word_name)) = parser.next() else { println!("ERROR: Word name not given."); std::process::exit(1); };
			    let (_, word_content) = paren_matching!(|x| matches!(x, ":"), |x| matches!(x, ";"));

			    self.dictionary.words.insert(word_name.to_string(), word_content);
			},
			";" => {},
			"BYE"|"bye" => {
			    std::process::exit(0);
			},
			"clearstack" => {
			    self.stack = Vec::new();
			},
			word => {
			    let words = self.dictionary.words.clone();
			    if let Some(word_content) = words.get(word) {
				self.vmrun(&mut Parser::new(word_content));
			    } else {
				println!("WARNING: word {:?} does not exist, will clear stack.", word);
				self.stack = Vec::new();
			    }
			}
		    }
		}
	    }
	}
    }
}
